﻿using System;
using System.Configuration;
using System.Diagnostics;
using System.Globalization;

namespace PicklistUpdate {
	internal static class Program {
		// app.config values
		private static string ServerType = ConfigurationManager.AppSettings["ServerType"];
		private static string ServerName = ConfigurationManager.AppSettings["ServerName"];
		private static string DatabaseName = ConfigurationManager.AppSettings["DatabaseName"];
		private static string DatabaseUserName = ConfigurationManager.AppSettings["DatabaseUserName"];
		private static string DatabasePassword = ConfigurationManager.AppSettings["DatabasePassword"];
		private static string SapUserName = ConfigurationManager.AppSettings["SapUserName"];
		private static string SapPassword = ConfigurationManager.AppSettings["SapPassword"];
		private static string Picklist = ConfigurationManager.AppSettings["Picklist"];
		private static string[] PickedRows = ConfigurationManager.AppSettings["picked"].Split('|');

		private static void Main(string[] args) {
			try {
				Console.WriteLine(" Connecting to company...");
				SAPbobsCOM.Company company = ConnectToCompany();
				Console.WriteLine();
				CheckParameters();
				Console.WriteLine("   Setting allocations in BaseDoc... ");
				UpdateConfiguredBaseDoc(company);
				Console.WriteLine("\n   Setting allocations in Picklist... ");
				UpdateConfiguredPickList(company);
				Console.WriteLine();
			} catch (Exception ex) {
				Console.WriteLine("   Error: " + ex.Message);
			}
			Console.ReadKey();
			Console.WriteLine();
			Console.Write(" Exiting... ");
		}

		private static void UpdateConfiguredBaseDoc(SAPbobsCOM.Company company) {
			Stopwatch stopwatchFull = Stopwatch.StartNew();
			// Get Sales Order
			SAPbobsCOM.Documents baseDocument = (SAPbobsCOM.Documents)company.GetBusinessObject(SAPbobsCOM.BoObjectTypes.oOrders);
			string salesOrderEntry = PickedRows[0].Replace("\r\n", string.Empty).Split(';')[6];
			baseDocument.GetByKey(int.Parse(salesOrderEntry));
			foreach (string pickedRow in PickedRows) {
				Stopwatch stopwatchInner = Stopwatch.StartNew();
				if (pickedRow.Length < 1) continue;
				// Get picking data
				string[] picked = pickedRow.Replace("\r\n", string.Empty).Split(';');
				double pickedQuantity = double.Parse(picked[2], CultureInfo.InvariantCulture);
				string batchNumber = picked[3];
				int salesOrderLine = int.Parse(picked[8]);
				// Fill DI properties
				for (int i = 0; i < baseDocument.Lines.Count; i++) {
					baseDocument.Lines.SetCurrentLine(i);
					if (baseDocument.Lines.LineNum != salesOrderLine) continue;
					SAPbobsCOM.BatchNumbers batchNums = baseDocument.Lines.BatchNumbers;
					bool batchFound = false;
					for (int j = 0; j < batchNums.Count; j++) {
						batchNums.SetCurrentLine(j);
						if (batchNums.BatchNumber != batchNumber) continue;
						batchNums.Quantity = pickedQuantity;
						batchFound = true;
					}
					if (!batchFound) {
						if (!string.IsNullOrEmpty(batchNums.BatchNumber)) batchNums.Add();
						batchNums.BatchNumber = batchNumber;
						batchNums.Quantity = pickedQuantity;
					}
					stopwatchInner.Stop();
					Console.WriteLine("      BaseDoc allocation (" + batchNumber + ") " + stopwatchInner.ElapsedMilliseconds + " ms");
				}
			}
			// DI Update
			Stopwatch stopwatchUpdate = Stopwatch.StartNew();
			int result = baseDocument.Update();
			stopwatchUpdate.Stop();
			if (result != 0) {
				string errMsg = company.GetLastErrorDescription();
				Console.WriteLine("   ERROR: " + result + " " + errMsg);
			}
			stopwatchFull.Stop();
			Console.WriteLine("   BaseDoc allocation total: " + stopwatchFull.ElapsedMilliseconds + " ms (update: " + stopwatchUpdate.ElapsedMilliseconds + " ms)");
		}

		private static void UpdateConfiguredPickList(SAPbobsCOM.Company company) {
			Stopwatch stopwatchFull = Stopwatch.StartNew();
			// Get Pick List
			SAPbobsCOM.PickLists pickList = (SAPbobsCOM.PickLists)company.GetBusinessObject(SAPbobsCOM.BoObjectTypes.oPickLists);
			pickList.GetByKey(int.Parse(Picklist));
			foreach (string pickedrow in PickedRows) {
				if (pickedrow.Length < 1) continue;
				// Get picking data
				string[] picked = pickedrow.Replace("\r\n", string.Empty).Split(';');
				double pickedQuantity = double.Parse(picked[2], CultureInfo.InvariantCulture);
				string batchNumber = picked[3];
				int binAbsEntry = int.Parse(picked[5]);
				int pickListLine = int.Parse(picked[7]);
				// Fill DI properies
				for (int i = 0; i < pickList.Lines.Count; i++) {
					Stopwatch stopwatchInner = Stopwatch.StartNew();
					pickList.Lines.SetCurrentLine(i);
					if (pickList.Lines.LineNumber != pickListLine) continue;
					// Set batch allocations
					SAPbobsCOM.BatchNumbers batchNums = pickList.Lines.BatchNumbers;
					bool batchFound = false;
					for (int j = 0; j < batchNums.Count; j++) {
						batchNums.SetCurrentLine(j);
						if (batchNums.BatchNumber != batchNumber) continue;
						batchNums.Quantity = pickedQuantity;
						batchFound = true;
					}
					if (!batchFound) {
						if (!string.IsNullOrEmpty(batchNums.BatchNumber)) batchNums.Add();
						batchNums.BatchNumber = batchNumber;
						batchNums.Quantity = pickedQuantity;
					}
					// Set bin allocations
					SAPbobsCOM.DocumentLinesBinAllocations binAllocs = pickList.Lines.BinAllocations;
					bool binFound = false;
					for (int j = 0; j < binAllocs.Count; j++) {
						binAllocs.SetCurrentLine(j);
						if (binAllocs.BinAbsEntry != binAbsEntry) continue;
						binAllocs.Quantity = pickedQuantity;
						binFound = true;
					}
					if (!binFound) {
						if (binAllocs.BinAbsEntry < 1) binAllocs.Add();
						binAllocs.BinAbsEntry = binAbsEntry;
						binAllocs.SerialAndBatchNumbersBaseLine = GetBatchBaseLine(batchNums, batchNumber);
						binAllocs.Quantity = pickedQuantity;
					}
					//pickList.Update();
					stopwatchInner.Stop();
					Console.WriteLine("      PickList allocation (" + batchNumber + ") " + stopwatchInner.ElapsedMilliseconds + " ms");
				}
			}
			// DI Update
			Stopwatch stopwatchUpdate = Stopwatch.StartNew();
			int result = pickList.Update();
			stopwatchUpdate.Stop();
			if (result != 0) {
				string errMsg = company.GetLastErrorDescription();
				Console.WriteLine("   ERROR: " + result + " " + errMsg);
			}
			stopwatchFull.Stop();
			Console.WriteLine("   PickList allocation total: " + stopwatchFull.ElapsedMilliseconds + " ms (update: " + stopwatchUpdate.ElapsedMilliseconds + " ms)");
		}

		private static int GetBatchBaseLine(SAPbobsCOM.BatchNumbers batchNums, string batchNumber) {
			for (int i = 0; i < batchNums.Count; i++) {
				batchNums.SetCurrentLine(i);
				if (batchNums.BatchNumber == batchNumber) return i;
			}
			return -1;
		}

		private static void CheckParameters() {
			if (string.IsNullOrEmpty(Picklist)) {
				throw new Exception("Picklist number must be set in app.config");
			}
		}

		private static SAPbobsCOM.Company ConnectToCompany() {
			SAPbobsCOM.Company company = new SAPbobsCOM.Company();
			company.Server = ServerName;
			company.CompanyDB = DatabaseName;
			company.UserName = SapUserName;
			company.Password = SapPassword;
			company.UseTrusted = false;
			switch (ServerType) {
				case "HANA": company.DbServerType = SAPbobsCOM.BoDataServerTypes.dst_HANADB; break;
				case "SQL2008": company.DbServerType = SAPbobsCOM.BoDataServerTypes.dst_MSSQL2008; break;
				case "SQL2012": company.DbServerType = SAPbobsCOM.BoDataServerTypes.dst_MSSQL2012; break;
				case "SQL2014": company.DbServerType = SAPbobsCOM.BoDataServerTypes.dst_MSSQL2014; break;
				case "SQL2016": company.DbServerType = SAPbobsCOM.BoDataServerTypes.dst_MSSQL2016; break;
				default: throw new Exception("Invalid ServerType '" + ServerType + "' in app.config");
			}
			company.DbUserName = DatabaseUserName;
			company.DbPassword = DatabasePassword;
			int res = company.Connect();
			if (res != 0) {
				int errCode;
				string errMsg;
				company.GetLastError(out errCode, out errMsg);
				throw new Exception(" Connection to company failed: " + errCode + " " + errMsg);
			}
			Console.WriteLine(" Connection to company successful.");
			return company;
		}
	}
}